﻿local debug	= false

CLootMasterML	      = LibStub("AceAddon-3.0"):NewAddon("CLootMasterML", "AceConsole-3.0", "AceComm-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0")
local CLootMaster    = LibStub("AceAddon-3.0"):GetAddon("CouncilLootMaster")

local L = LibStub("AceLocale-3.0"):GetLocale("CouncilLootMaster")
local LibDialog = LibStub("LibDialog-1.0")

local addon         = CLootMasterML		-- Local instance of the addon

local MsgPrefix     = 'CouncilLootmaster: '

local rosterNew = true

local currentEnchNeeded = 575

local groupSize = 0


-- Using LibDialog to prevent UI Taints
LibDialog:Register("COUNCILLOOTMASTER_ASK_TRACKING", {
	text = '- - - - CouncilLootmaster - - - -\r\n\r\nYou are the loot master, would you like to use CouncilLootmaster to distribute loot?\r\n\r\n(You will be asked again next time. Use /cloot config to change this behaviour)',

	buttons = {
		{ 	text = YES,
			on_click = function(self)
				CLootMasterML:EnableTracking()
				CLootMaster:Print(L['You have enabled loot tracking for this raid'])
			end,
		},
		{ 	text = NO,
			on_click = function()
				CLootMasterML:DisableTracking()
				CLootMaster:Print(L['You have disabled loot tracking for this raid'])
			end},
	},

	hide_on_escape = false,
	show_while_dead = true,
	exclusive = true,
})

function CLootMasterML:Debug( message, verbose )
	if not CLootMaster or not CLootMaster.debug then return end
	if verbose and not CLootMaster.verbose then return end
	self:Print("debug: " .. message)
end

--[[
	Initialization function
]]
function CLootMasterML:OnInitialize()

	self.lootTable = {}			-- a table to store loot info

	self.enchanterTable = {
		['enchanters'] = {},
		['rowdata'] = {}
	}	-- a table to store enchanters

	self.currentDE = nil

	-- Event Register
	-- Trap event when ML rightclicks master loot
	self:RegisterEvent("OPEN_MASTER_LOOT_LIST")

	-- Trap event when an item gets looted
	self:RegisterEvent("CHAT_MSG_LOOT")

	-- Trap some important system messages here
	self:RegisterEvent("GROUP_ROSTER_UPDATE",            "GROUP_UPDATE")
	self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED",     "GROUP_UPDATE")
	self:RegisterEvent("PLAYER_ENTERING_WORLD",         "GROUP_UPDATE")

	-- Trap events when entering and leaving combat
	self:RegisterEvent("PLAYER_REGEN_DISABLED",     "EnterCombat")
	self:RegisterEvent("PLAYER_REGEN_ENABLED",      "LeaveCombat")

	self:RegisterEvent("GUILD_ROSTER_UPDATE",       "CacheGuildInfo")

	-- Trap event when a loot window is opened
	self:RegisterEvent("LOOT_OPENED")

	-- Register communications
	self:RegisterComm("CLootMasterML", 		    "CommandReceived")

	-- Create table for the guildinfo cache.
	self.guildInfo = {}

	-- Change the onClick script of the lootbuttons a little so we can trap alt+clicks
	-- NOTE: Only tested with normal wow lootframes, not using XLoot etc.
	for slot=1, LOOTFRAME_NUMBUTTONS do
		local btn = getglobal("LootButton"..slot)
		if btn and not btn.oldClickEventCouncilLM then
			btn.oldClickEventCouncilLM = btn:GetScript("OnClick")
			btn:SetScript("OnClick", function(btnObj, ...)
				if not IsAltKeyDown() then
					return btnObj.oldClickEventCouncilLM(btnObj, ...)
				end
				return LootButton_OnClick(btnObj, ...)
			end)
		end
	end

	-- Just to be sure, enable the tracking.
	self:EnableTracking()
end

--[[
	Enable loot tracking
]]
function CLootMasterML:EnableTracking()
	self.trackingEnabled = true
end

--[[
	Disable loot tracking
]]
function CLootMasterML:DisableTracking()
	self.trackingEnabled = false
end

--[[
	Check whether tracking is enabled
]]
function CLootMasterML:TrackingEnabled()
	return self.trackingEnabled
end

--[[
	Returns current version string
]]
function CLootMasterML:GetVersionString()
	local lm = LibStub("AceAddon-3.0"):GetAddon("CouncilLootMaster")
	return (lm.GetVersionString(lm) or 'unknown version')
end

--[[
	Sends a response to player by in-game chat whisper
]]
function CLootMasterML:SendWhisperResponse(message, target)
	SendChatMessage( MsgPrefix .. ( message or ''), 'WHISPER', nil, target )
	return false
end

--[[
	Send commands to RAID/PARTY/GUILD/PLAYERS through addon channels
]]
function CLootMasterML:SendCommand(command, message, target)
	if not target then
		return self:Print(L["Could not send command, no target specified"])
	end
	if target=='RAID' then
		self:SendCommMessage("CLootMasterC", format("%s:%s", tostring(command), tostring(message)), "RAID", nil, "ALERT")
	elseif target=='PARTY' then
		self:SendCommMessage("CLootMasterC", format("%s:%s", tostring(command), tostring(message)), "PARTY", nil, "ALERT")
	elseif target=='GUILD' then
		self:SendCommMessage("CLootMasterC", format("%s:%s", tostring(command), tostring(message)), "GUILD", nil, "ALERT")
	else
		-- Don't use AceComm for messages to self, call function directly
		if target==UnitName('player') then
			CLootMaster.CommandReceived(CLootMaster, "CLootMasterC", format("%s:%s", tostring(command), tostring(message)), 'WHISPER', target)
		else
			self:SendCommMessage("CLootMasterC", format("%s:%s", tostring(command), tostring(message)), "WHISPER", target, "ALERT")
		end
	end
end

--[[
	Function to parse a monitor message
]]
function CLootMasterML:ParseMonitorMessage( message )
	local tpl = {}
	for v in string.gmatch(message, '(.-)(%^%*)') do
		if v=='nil' then v=nil end
		tinsert(tpl, v)
	end
	for _,_,v in string.gmatch(message, '(.*)(%^%*)(.-)$') do
		if v=='nil' then v=nil end
		tinsert(tpl, v)
	end
	if #tpl==0 then
		if message=='nil' then message=nil end
		tinsert(tpl, message)
	end

	return tpl
end

-- Table linking message priorities to somewhat unintuitive priority names in AceComm
local MonitorMessagePriorities = {
	["PRIORITY_HIGH"]       = "ALERT",
	["PRIORITY_NORMAL"]     = "NORMAL",
	["PRIORITY_LOW"]        = "BULK"
}

--[[
	Function to check whether monitor messages are required for this piece of loot
]]
function CLootMasterML:MonitorMessageRequired( loot )
	if not CLootMaster.db.profile.monitorSend then return false end
	loot = self:GetLoot( loot )
	if not loot then return false end
	if not loot.mayDistribute then return false end
	if loot.autoLootable then return false end
	if loot.binding == 'pickup' then return true end
	if loot.quality>=CLootMaster.db.profile.monitorThreshold then return true end
	return false
end

--[[
	Sends a message to enabled monitors, only sends ADDLOOT messages by whisper,
	the rest will get ignored if there has been no ADDLOOT message, and this
	allows use of the RAID channel rather than WHISPER
]]
function CLootMasterML:SendMonitorMessage(...)

	if not CLootMaster.db.profile.monitorSend then return end

	local numArgs = select("#", ...)

	local prio = nil
	if numArgs>0 then
		prio = select(numArgs, ...)
		prio = MonitorMessagePriorities[prio]
	end

	local args = {...}

	if prio then
		tremove(args)
	else
		prio = "NORMAL"
	end
	
	self:SendMonitorMessageEx(prio, unpack(args))
end

--[[
	Does the actual monitor message sending based on monitor settings
]]
function CLootMasterML:SendMonitorMessageEx(prio, ...)
	
	if not CLootMaster.db.profile.monitorSend then return end
	local numArgs = select("#", ...)

	local out = ""
	for i=1,numArgs do
		if i > 1 then
			out = out .. "^*"
		end
		out = out .. tostring(select(i, ...))
	end

	local msgType = tostring(select(1, ...))
	
	local num = GetNumGroupMembers()
	if num>0 then
		-- we're in raid
		if msgType=="ADDLOOT" then
			local monitorPromoted = CLootMaster.db.profile.monitorSendAssistantOnly
			local monitorSpecificRanks = CLootMaster.db.profile.monitorSendGuildRank
			local monitorRankList = CLootMaster.db.profile.monitorSendGuildRankList
			local votingPromoted = CLootMaster.db.profile.votingSendAssistantOnly
			local votingSpecificRanks = CLootMaster.db.profile.votingSendGuildRank
			local votingRankList = CLootMaster.db.profile.votingSendGuildRankList
			
			if (monitorPromoted or monitorSpecificRanks or votingPromoted or votingSpecificRanks) then
				-- Only send to promoted players or specific guildranks in raid, use whispers.
				for raidIndex=1, MAX_RAID_MEMBERS do
					local rName, rRank, _, _, _, _, _, rOnline, _, _, rIsML = GetRaidRosterInfo(raidIndex)

					if rName and rOnline then
						local sendMessage = monitorPromoted and (rRank>0 or rIsML)
						local isVoter = votingPromoted and (rRank>0 or rIsML)
						-- Send to specific guildrank
						if not (sendMessage or isVoter) and UnitIsInMyGuild(rName) and (monitorSpecificRanks or votingSpecificRanks) then
							local _, _, guildRank = GetGuildInfo(rName)
							if monitorSpecificRanks and type(monitorRankList)=='table' then
								sendMessage = monitorRankList[guildRank+1]
							end
							if votingSpecificRanks and type(votingRankList)=='table' then
								isVoter = votingRankList[guildRank+1]
							end
						end
						if not (monitorPromoted or monitorSpecificRanks) then
							sendMessage = true
						end
						if not (votingPromoted or votingSpecificRanks) then
							isVoter = true
						end
						
						-- Add vote status
						if isVoter then
							out = out .. "^*dovote"
						elseif CLootMaster.db.profile.votingWatchAllowed then
							out = out .. "^*watchvote"
						else
							out = out .. "^*hidevote"
						end
						
						if sendMessage and rName~=UnitName('player') then
							-- Player is online and eligible for monitoring
							self:SendCommMessage("CLootMasterML", format("MONITOR:%s", out), "WHISPER", rName, prio)
						end
					end
				end
			else
				-- Everybody is allowed to monitor, just use the raid channel
				self:SendCommMessage("CLootMasterML", format("MONITOR:%s", out), "RAID", nil, prio)
			end
		else
			-- Everybody is allowed to monitor, just use the raid channel
			self:SendCommMessage("CLootMasterML", format("MONITOR:%s", out), "RAID", nil, prio)
		end
	else
		num = GetNumSubgroupMembers()
		if num>0 then
			if msgType=="ADDLOOT" then
				local monitorSpecificRanks = CLootMaster.db.profile.monitorSendGuildRank
				local monitorRankList = CLootMaster.db.profile.monitorSendGuildRankList
				local votingSpecificRanks = CLootMaster.db.profile.votingSendGuildRank
				local votingRankList = CLootMaster.db.profile.votingSendGuildRankList
			
				if monitorSpecificRanks or votingSpecificRanks then
					-- Only send to specific guildranks in party, use whispers.
					for partyIndex=1, MAX_PARTY_MEMBERS do
						if GetPartyMember(partyIndex) then
							local unit = 'party' .. partyIndex
							local unitName = UnitName(unit)
							local unitOnline = UnitIsConnected(unit)
							local unitInGuild = UnitIsInMyGuild(unit)
							local _, _, guildRank = GetGuildInfo(unitName)
							print(unit, unitName, unitOnline, unitInGuild, guildRank)
							if unitName and guildRank and not UnitIsUnit(unitName,'player') and unitOnline and unitInGuild and ((not monitorSpecificRanks) or (type(monitorRankList)=='table' and monitorRankList[guildRank+1]))then
								-- Player is online and not self and a monitor rank
								local isVoter = (not votingSpecificRanks) or (type(votingRankList)=='table' and votingRankList[guildRank+1])
								
								-- Add vote status
								if isVoter then
									out = out .. "^*dovote"
								elseif CLootMaster.db.profile.votingWatchAllowed then
									out = out .. "^*watchvote"
								else
									out = out .. "^*hidevote"
								end
								
								self:Debug('SendMonitorMessage('..unitName..'): ' .. out, true)
								self:SendCommMessage("CLootMasterML", format("MONITOR:%s", out), "WHISPER", unitName, prio)
							end
						end
					end
				else
					--we're in party
					self:SendCommMessage("CLootMasterML", format("MONITOR:%s", out), "PARTY", nil, prio)
					self:Debug('SendMonitorMessage(PARTY): ' .. out, true)
				end
			else
				--we're in party
				self:SendCommMessage("CLootMasterML", format("MONITOR:%s", out), "PARTY", nil, prio)
				self:Debug('SendMonitorMessage(PARTY): ' .. out, true)
			end
		else
			--we're not grouped, send message to self for debugging purposes.
			self:SendCommMessage("CLootMasterML", format("MONITOR:%s", out), "WHISPER", UnitName('player'), prio)
			self:Debug('SendMonitorMessage(WHISPER->SELF): ' .. out, true)
		end
	end
end

--[[
	Event gets triggered when ML receives a message from a candidate
]]
function CLootMasterML:CommandReceived(prefix, message, distribution, sender)

	local _,_,command, message = string.find(message, "^([%a_]-):(.*)$")
	command = strupper(command or '')
	message = message or ''

	if command == 'WANT' then

		-- A candidate just told us he wants the item
		local itemID, response, avgItemLvl, note = strsplit('^', message)
		avgItemLvl = tonumber(avgItemLvl)
		response = tonumber(response)
		note = note or ''

		local loot = self:GetLoot( itemID )
		if loot and ((loot.mayDistribute and self:IsCandidate(itemID, sender)) or not loot.mayDistribute) then

			if not loot.mayDistribute and loot.candidatesReceived and not self:IsCandidate(itemID, sender) then
				-- we're monitoring, candidatelist has been received but the responding player is not
				-- on the candidatelist, ignore
				return
			end

			if CLootMaster.db.profile.ignoreResponseCorrections and tonumber(self:GetCandidateData( itemID, sender, 'response' ) or 0) >= CLootMaster.RESPONSE.MAIN then
				self:Debug( format("%s responded WANT for %s but there already is a response and ignoreResponseCorrections is enabled.", sender or 'nil', itemID or 'nil'))
			else
				self:SetCandidateResponse( loot.id, sender, response )
			end
			self:SetCandidateData( loot.id, sender, 'note', note )
			self:SetCandidateData( loot.id, sender, 'avgItemLvl', avgItemLvl )
			self:ReloadMLTableForLoot( loot.id )
		else
			self:Debug( format("%s responded WANT for %s but loot not found (Already looted?)", sender or 'nil', itemID or 'nil'))
		end

	elseif command == 'GEAR' then

		-- A candidate just sent us his current gear. update the cache
		local _,_,itemID, gear = string.find(message, "^([^\^]-)\^(.*)$")
		local item1, item2 = strsplit('$', gear)

		local loot = self:GetLoot( itemID )

		if loot and ((loot.mayDistribute and self:IsCandidate(itemID, sender)) or not loot.mayDistribute) then

			if not loot.mayDistribute and loot.candidatesReceived and not self:IsCandidate(itemID, sender) then
				-- we're monitoring, candidatelist has been received but the responding player is not
				-- on the candidatelist, ignore
				return
			end

			self:SetCandidateData( loot.id, sender, 'currentitem', item1, true )
			self:SetCandidateData( loot.id, sender, 'currentitem2', item2, true )
			self:SetCandidateData( loot.id, sender, 'foundGear', true, true )
			self:SetCandidateResponse( loot.id, sender, CLootMaster.RESPONSE.WAIT )
			self:ReloadMLTableForLoot( loot.id )
		else
			self:Debug( format("%s responded GEAR for %s but loot or candidate not found (Already looted?)", sender or 'nil', itemID or 'nil'))
		end

	elseif command == 'VOTE' then

		-- A member of the loot council sent the master looter a vote for a specific item
		local itemID, candidate = strsplit('^', message)
		local loot = self:GetLoot( itemID )
		
		-- loot not found
		if not loot then
			self:Debug( format("%s responded VOTE for %s but loot not found (Already looted?)", sender or 'nil', itemID or 'nil'))
			return
		end
		
		-- not allowed to watch votes
		if not loot.voteWatching then
			return
		end

		-- filter out duplicate votes to self...
		if UnitIsUnit(sender, 'player') and prefix~=nil and loot.mayDistribute then return end

		-- You've voted, another vote is nolonger required
		if UnitIsUnit(sender, 'player') then
			loot.voteRequired = false
		end

		if not self:IsCandidate(loot.id, candidate) then
			print(candidate .. ' is no candidate for ' .. itemID .. '. (Sent by ' .. sender .. ')')
			return
		end
		
		if type(loot.voters) ~= 'table' then
			-- noone ever voted, create the table
			loot.voters = {}
		elseif type(loot.voters[sender]) == 'string' then
			-- sender already voted, unvote
			self:SetCandidateData(loot.id, loot.voters[sender], 'votes', (tonumber(self:GetCandidateData(loot.id, loot.voters[sender], 'votes')) or 0)-1, true)
		end
		
		loot.voters[sender] = candidate
		self:SetCandidateData(loot.id, candidate, 'votes', (tonumber(self:GetCandidateData(loot.id, candidate, 'votes')) or 0)+1, true)
		self:ReloadMLTableForLoot( loot.id )

	elseif command == 'ENCHRESP' then

		if message then
			self:AddEnchanter(sender)
		end

	elseif command == 'MONITOR' then

		-- Only handle monitor messages if we enabled them in the options screen.
		if not CLootMaster.db.profile.monitor then return end

		-- ignore our own monitor messages!
		if sender == UnitName('player') then return end

		local monArgs = self:ParseMonitorMessage(message)
		local monCmd = tremove(monArgs, 1)

		if monCmd == 'ADDLOOT' then
			local itemLink, itemName, itemID, ilevel, itemBind, itemRarity, itemTexture, itemEquipLoc, quantity, voteStatus = unpack(monArgs)

			if not self.lootTable then self.lootTable={} end

			if self.lootTable[itemID] then
				-- Is someone tinkering? the loot already exists. Return
				return
			end

			itemRarity = tonumber(itemRarity) or 0

			-- Check if the loot < monitorIncomingThreshold; cancel
			if itemRarity < CLootMaster.db.profile.monitorIncomingThreshold then
				-- Do not show monitor window for this item.
				return
			end
			
			local voteRequired = (voteStatus=="dovote")
			local voteWatching = not (voteStatus=="hidevote")

			self.lootTable[itemID] = {
				['link']	                    = itemLink,
				['name']	                    = itemName,

				['lootmaster']                  = sender,

				['announced']                   = true,
				['mayDistribute']               = false,
				['voteRequired']				= voteRequired,
				['voteWatching']				= voteWatching,

				['id']                          = itemID,
				['itemID']                      = itemID,
				['itemid']                      = itemID,

				['ilevel']	                    = tonumber(ilevel) or 0,
				['binding']	                    = itemBind,
				['quality']                     = itemRarity,
				['quantity']                    = tonumber(quantity) or 1,

				['votes']	                    = 0,

				['texture']                     = itemTexture or '',

				['equipLoc']	                = itemEquipLoc or '',

				['started']	                    = nil,

				['rowdata']	                    = {},
				['candidates']	                = {},
				['numResponses']                = 0
			}

			self:ReloadMLTableForLoot( itemID )

		elseif monCmd == 'CANDIDATELIST' then

			-- in an attempt to speed the monitors up, lets just receive all candidate data for new loot in big chunk
			-- not candidate by candidate, updates hereafter are still one by one.

			local mlCandidates = {}
			local itemID, fieldCount, candidates

			if monCmd == 'CANDIDATELIST' then
				itemID, fieldCount, candidates = unpack(monArgs)
				fieldCount = tonumber(fieldCount) or 1
			end

			local loot = self:GetLoot(itemID)
			if not loot or loot.mayDistribute or loot.lootmaster~=sender then
				-- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
				return
			end

			-- Don't allow players to be added unless they are already on the list.
			loot.candidatesReceived = true

			if candidates then
				local cdata = { strsplit( ' ', candidates ) }
				for i=1, fieldCount do
					self:AddCandidate( itemID, cdata[i] )
				end
			end

			-- Reload the UI
			self:ReloadMLTableForLoot( itemID )

		elseif monCmd == 'SETCANDIDATEDATA' then

			local itemID, candidate, name, value = unpack(monArgs)
			local loot = self:GetLoot(itemID)

			if not loot or loot.mayDistribute or loot.lootmaster~=sender  then
				-- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
				return
			end

			if not self:IsCandidate( itemID, candidate ) then return end
			self:SetCandidateData( itemID, candidate, name, value )

			self:ReloadMLTableForLoot( itemID )

		elseif monCmd == 'CANDIDATEGEAR' then

			local itemID, candidate, item1, item2 = unpack(monArgs)
			local loot = self:GetLoot(itemID)

			if not loot or loot.mayDistribute or loot.lootmaster~=sender  then
				-- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
				return
			end

			if not self:IsCandidate( itemID, candidate ) then return end

			if item1 then
				self:SetCandidateData( itemID, candidate, 'currentitem', item1 )
			end
			if item2 then
				self:SetCandidateData( itemID, candidate, 'currentitem2', item2 )
			end

			self:SetCandidateData( itemID, candidate, 'foundGear', true, true )

			self:ReloadMLTableForLoot( itemID )

		elseif monCmd == 'REMOVELOOT' or monCmd == 'DECREASELOOT' then

			local itemID = unpack(monArgs)
			local loot = self:GetLoot(itemID)

			if not loot or loot.mayDistribute or loot.lootmaster~=sender then
				-- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
				return
			end

			self:RemoveLoot(itemID)
			self:UpdateUI()

		end
	end
end

--[[
	Send a request to the specified player if he/she is interested in the given loot
	candidate could also be RAID or PARTY
]]
function CLootMasterML:AskCandidateIfNeeded( link, candidate )

	local loot = self:GetLoot( link )
	if not loot then
		return self:Print( format(L["Could not ask player if needed because %s is not cached"], link) )
	end

	-- Hardcoded time in secs until autopass
	self.timeout = CLootMaster.db.profile.loot_timeout

	local notesAllowed = 0
	if CLootMaster.db.profile.allowCandidateNotes then
		notesAllowed = 1
	end

	self:SendCommand( "DO_YOU_WANT", format('%s^%s^%s^%s^%s^%s^%s^%s^%s^%d',
											loot.id or '',
											loot.ilevel or -1,
											loot.binding or '',
											loot.equipLoc or 0,
											loot.quality or 0,
											self.timeout or 60,
											loot.link,
											loot.texture or '',
											notesAllowed,
											loot.encodedClasses or CLootMaster:FullClassesEncoded()
											), candidate )

	-- Update the ui
	self:ReloadMLTableForLoot( loot.id )
end

--[[
	Add loot to masterloot cache. This is where candidate responses are stored, plus the rows for the scrollingtable
]]
function CLootMasterML:AddLoot( link, mayDistribute, quantity )
	if not link then return end
	if not self.lootTable then self.lootTable={} end

	if self.lootTable[link] then return link end

	local itemName, itemLink, _, _, itemMinLevel, itemType, itemSubType, itemStackCount, _, itemTexture = GetItemInfo(link)

	local _,_,itemID = strfind(itemLink, 'Hitem:(%d+)')
	if not itemID or not itemName then return end

	if self.lootTable[itemID] then return itemID end

	local db = CLootMaster.db.profile

	-- Get the itemEquipLocation, rarity and level, and keep Tier Tokens in consideration.
	local itemEquipLoc, itemRarity, ilevel = GetCustomSlotInfo(link)

	-- See if the item is BoP, BoE or BoU
	local itemBind = CLootMaster:GetItemBinding(itemLink)

	self.lootTable[itemID] = {
		['link']	          = itemLink,
		['name']	          = itemName,

		['announced']       = true,
		['mayDistribute']   = mayDistribute,
		['voteRequired']	= true,
		['voteWatching']	= true,

		['id']              = itemID,
		['itemID']          = itemID,
		['itemid']          = itemID,

		['ilevel']	        = ilevel or 0,
		['binding']	        = itemBind or '',
		['quality']         = itemRarity or 0,
		['quantity']        = quantity or 1,

		['votes']	        = 0,

		['texture']         = itemTexture or '',

		['equipLoc']	    = itemEquipLoc or '',

		['started']	        = nil,

		['rowdata']	        = {},
		['candidates']	    = {},

		['numResponses']        = 0
	}
	local loot = self.lootTable[itemID]

	-- See if this item should be autolooted
	if db.AutoLootThreshold~=0 and db.AutoLooter~='' then
		if (not itemBind or itemBind=='use' or itemBind=='equip') and itemRarity<=db.AutoLootThreshold then
			loot.autoLootable = true
		end
	end

	-- Are we lootmaster for this loot? Lets send out a monitor message about the added loot
	if loot.mayDistribute and self:MonitorMessageRequired(itemID) then
		self:SendMonitorMessage('ADDLOOT', itemLink, itemName, itemID, ilevel or 0, itemBind or '', itemRarity or 0, itemTexture or '', itemEquipLoc or '', quantity or 1, "PRIORITY_HIGH" )
	end

	return itemID
end

--[[
	Casts vote for a given item to a given candidate by sending to lootmaster
]]
function CLootMasterML:CastVote( link, candidate )
	local loot = self:GetLoot(link)
	if not loot then return end
	loot.voteRequired = false
	CLootMaster:SendCommand('VOTE', format('%s^%s', loot.id, candidate), loot.lootmaster or UnitName('player'))

	self:ReloadMLTableForLoot( loot.id )
end

--[[
	Announce a piece of loot to the raid
]]
function CLootMasterML:AnnounceLoot(loot)
	loot = self:GetLoot( loot )
	if not loot then return end

	if loot.announced == true then return end

	loot.announced = true

	-- Also send a text message and an addon message to the raid/party chat.
	if CLootMaster:UnitInRaid('player') then
		-- we're in raid with master looter
		self:AskCandidateIfNeeded( loot.id, 'RAID' )
	elseif UnitInParty('player') and GetNumSubgroupMembers()>0 then
		--we're in party with master looter
		self:AskCandidateIfNeeded( loot.id, 'PARTY' )
	end

	-- Traverse the candidate list, see if there is anyone eligible for the loot, but not grouped anymore...
	for candidate, index in pairs(loot.candidates) do
		if not CLootMaster:UnitInRaid(candidate) and not (UnitInParty(candidate) and GetNumSubgroupMembers()>0) and tonumber(self:GetCandidateData( loot.id, candidate, 'response')) == CLootMaster.RESPONSE.NOTANNOUNCED then
			self:AskCandidateIfNeeded( loot.id, candidate )
		end
	end
end


--[[
	Get the lootdata from masterloot cache. This is where candidate responses are stored, plus the rows for the scrollingtable. Returns nil if not found
]]
function CLootMasterML:GetLoot( link )
	if not link then self:Debug('getloot: !link'); return nil end
	if not self.lootTable then self:Debug('getloot: !lootTable'); return nil end

	if self.lootTable[link] then return self.lootTable[link] end

	local _,_,itemID = strfind(link, 'Hitem:(%d+)')
	if not itemID then self:Debug('getloot: !itemID: ' .. tostring(itemID)); return end
	if not self.lootTable[itemID] then self:Debug('getloot: !lootTable[itemID]: ' .. tostring(itemID)); return end

	return self.lootTable[itemID]
end

--[[
	Get the lootdata from masterloot cache. This is where candidate responses are stored, plus the rows for the scrollingtable. Returns nil if not found
]]
function CLootMasterML:GetLootID( link )
	if not link then self:Debug('getlootid: !link'); return nil end
	if not self.lootTable then self:Debug('getlootid: !lootTable'); return nil end

	if self.lootTable[link] then return link end

	local _,_,itemID = strfind(link, 'Hitem:(%d+)')
	if not itemID then self:Debug('getlootid: !itemID: ' .. tostring(itemID)); return end
	if not self.lootTable[itemID] then self:Debug('getloot: !lootTable[itemID]: ' .. tostring(itemID)); return end

	return itemID
end

--[[
	Returns an itemID from a given itemLink, even if the item has not been seen before (WoW limitation on GetItemInfo())
]]
function CLootMasterML:GetItemIDFromLink( link )
	if not link then return nil end
	local _,_,itemID = strfind(link, 'Hitem:(%d+)')
	if not itemID then return nil end
	return tonumber(itemID)
end

--[[
	Removes the given loot from the masterloot cache
]]
function CLootMasterML:RemoveLoot( link )
	if not link then return end
	if not self.lootTable then return end

	local loot = self:GetLoot(link)
	local itemID = loot.id

	if not itemID or not self.lootTable[itemID] then
		return self:Debug(format('RemoveLoot: not found %s %s(%s)', link or 'nil', itemID or 'nil', type(itemID)))
	end

	-- we have more than one of this item, decrease counter and return.
	if loot.quantity>1 then
		loot.quantity = loot.quantity - 1

		-- Are we lootmaster for this loot? Lets send out a monitor message about the quantity decrease
		if self.lootTable[itemID].mayDistribute and self:MonitorMessageRequired(itemID) then
			self:SendMonitorMessage( 'DECREASELOOT', itemID, "PRIORITY_HIGH" )
		end

		return true
	end

	if self.frame and self.frame.currentLoot and self.frame.currentLoot.itemID == itemID then
		self.frame.currentLoot = nil
	end

	self:Debug('RemoveLoot: '..link)

	-- Are we lootmaster for this loot? Lets send out a monitor message about the removed loot
	if self.lootTable[itemID].mayDistribute and self:MonitorMessageRequired(itemID) then
		self:SendMonitorMessage( 'REMOVELOOT', itemID, "PRIORITY_HIGH" )
	end

	self.lootTable[itemID] = nil
	return true
end

--[[
	Returns the number of votes for a given candidate for a given item
]]
function CLootMasterML:GetVoteValue(candidate, itemID)
	local loot = self:GetLoot(itemID)
	if loot.voteRequired then return 'vote' end
	return tonumber(self:GetCandidateData(loot.id, candidate, "votes")) or 0
end

--[[
	Returns the sorting value of a given candidates response to a given item
]]
function CLootMasterML:ResponseSortValue(candidate, itemID)
	local data = self:GetLoot(itemID)
	local response = tonumber(self:GetCandidateData(itemID, candidate, "response"))
	return self:GetCandidateData( itemID, candidate, 'responsesort')
end

--[[
	Adds a candidate (player eligible to receive loot) to the masterloot cache
	for the given loot
]]
function CLootMasterML:AddCandidate( loot, candidate )
	local itemID = self:AddLoot( loot )
	if not itemID then return end
	if not candidate then return end
	if not self.lootTable[itemID] then return end

	-- Just return the itemID if the candidate already exists
	if self.lootTable[itemID]['candidates'][candidate] then return itemID end

	-- Find the guildRank and unlocalized classname from the guildCache
	local guildRankName = ''
	local candidateClass = nil
	local candidateClassLocalized = nil
	if self.guildInfo and self.guildInfo[candidate] then
		guildRankName = self.guildInfo[candidate].rank or ''
		candidateClass = self.guildInfo[candidate].class
		candidateClassLocalized = self.guildInfo[candidate].classLocalized
	end

	-- No class found, try looking it up another way.
	if not candidateClass then
		candidateClassLocalized, candidateClass = UnitClass(candidate)
		-- Update the reverse class lookup table
		if candidateClassLocalized and candidateClass then
			CLootMaster:UpdateClassLocalizer(candidateClassLocalized, candidateClass)
		end
	end

	local initResponse = CLootMaster.RESPONSE.NOTANNOUNCED

	local initResponseSort = CLootMaster.RESPONSE[initResponse].SORT

	tinsert( self.lootTable[itemID]['rowdata'],
	{   ["unitclass"]       = candidateClass,
		["response"]        = initResponse,
		["responsesort"]    = initResponseSort,
		["onclick"]         = function(arg1, button, down) if button=='RightButton' then self:OnCandidateRowRightClick(candidate, loot, arg1) end end,
		["cols"] = {
			{["name"]       = 'class',
			["value"]      = candidateClass or '',
			["onenter"]    = addon.ShowCandidateCellPopup,
			["onleave"]    = addon.HideGearCellPopup,
			["userDraw"]   = addon.SetClassIconCellOwnerDraw,
			["onenterargs"]= { self, candidate, itemID },
			["onleaveargs"]= { self, candidate, itemID }
			},

			{["name"]       = 'name',
			["value"]      = candidate,
			["onenter"]    = addon.ShowCandidateCellPopup,
			["onleave"]    = addon.HideGearCellPopup,
			["color"]      = self.GetCandidateClassCellColor,
			["colorargs"]  = { self, candidate, itemID},
			["onenterargs"]= { self, candidate, itemID },
			["onleaveargs"]= { self, candidate, itemID }
			},

			{["name"]       = 'guildrank',
			["value"]      = guildRankName,
			["color"]      = self.GetCandidateCellColor,
			["colorargs"]  = {self, candidate, itemID}},

			{["name"]       = 'responsesortFunc',
			["value"]      = self.ResponseSortValue,
			["args"]       = {self, candidate, itemID},
			["userDraw"]   = addon.SetCandidateResponseCellUserDraw,
			["color"]      = self.GetCandidateCellColor,
			["colorargs"]  = {self, candidate, itemID}},

			{["name"]       = 'voteCol',
			["value"]      = self.GetVoteValue,
			["userDraw"]   = addon.SetCandidateVotesCellUserDraw,
			["onenter"]    = addon.ShowVotesCellPopup,
			["onleave"]    = addon.HideGearCellPopup,
			["onenterargs"]= {self,candidate,itemID},
			["onleaveargs"]= {self,candidate,itemID},
			["args"]       = {self,candidate,itemID}},


			{["name"]       = 'note',
			["value"]      = '',
			["userDraw"]   = addon.SetNoteCellOwnerDraw,
			["onenter"]    = addon.ShowNoteCellPopup,
			["onleave"]    = addon.HideInfoPopup,
			["onenterargs"]= { self,candidate,itemID,'currentitem' },
			["onleaveargs"]= { self,candidate,itemID,'currentitem' }
			},

			{["name"]       = ' ',
			["value"]      = ' ',
			["userDraw"]   = self.EmptyCellOwnerDraw},

			{["name"]       = 'currentilevel',
			["value"]      = '',
			["args"]       = {self, candidate, itemID},
			["onclick"]    = addon.OnGearInspectClick,
			["onenter"]    = addon.ShowGearInspectPopup,
			["onleave"]    = addon.HideGearCellPopup,
			["userDraw"]   = addon.SetGearCelliLVL,
			["onenterargs"]= { self,candidate,itemID },
			["onleaveargs"]= { self,candidate,itemID },
			["onclickargs"]= { self,candidate,itemID }
			},

			{["name"]       = 'currentitem',
			["value"]      = '',
			["userDraw"]   = addon.SetGearCellOwnerDraw,
			["onenter"]    = addon.ShowGearCellPopup,
			["onleave"]    = addon.HideGearCellPopup,
			["onclick"]    = addon.OnGearCellClick,
			["onenterargs"]= { self,candidate,itemID,'currentitem' },
			["onleaveargs"]= { self,candidate,itemID,'currentitem' },
			["onclickargs"]= { self,candidate,itemID,'currentitem' }
			},

			{["name"]       = 'currentitem2',
			["value"]      = '',
			["userDraw"]   = addon.SetGearCellOwnerDraw,
			["onenter"]    = addon.ShowGearCellPopup,
			["onleave"]    = addon.HideGearCellPopup,
			["onclick"]    = addon.OnGearCellClick,
			["onenterargs"]= { self,candidate,itemID,'currentitem2' },
			["onleaveargs"]= { self,candidate,itemID,'currentitem2' },
			["onclickargs"]= { self,candidate,itemID,'currentitem2' }
			},

			{["value"]      = ' '},     -- spacer
		}
	})

	self.lootTable[itemID]['candidates'][candidate] = #(self.lootTable[itemID]['rowdata'])

	self:InspectCandidate( itemID, candidate )

	return itemID
end

--[[
	Makes a list of current candidates for an item and sends a monitor message about it
]]
function CLootMasterML:SendCandidateListToMonitors( itemID )
	local loot = self:GetLoot( itemID )
	if not loot then return nil end

	local candidata = {}
	local numcandidates = 0
	for candidate, cIndex in pairs(loot.candidates) do
		numcandidates = numcandidates+1
		tinsert( candidata, candidate )
	end

	loot.candidatesSent = true
	self:SendMonitorMessage( 'CANDIDATELIST', itemID, numcandidates, strjoin(' ', unpack(candidata)) , 'PRIORITY_HIGH' )
end

--[[
	Returns true if the candidate already has been cached for the given loot
]]
function CLootMasterML:IsCandidate( loot, candidate )

	if not loot then return false end

	loot = self:GetLoot(loot)
	if not loot or not loot.id then return false end

	if not loot or not self.lootTable or not self.lootTable[loot.id] or not self.lootTable[loot.id]['candidates'] then return false end

	local rowID = self.lootTable[loot.id]['candidates'][candidate]
	if not rowID then return false end

	return true
end

--[[
	Stores some data in the candidate cache for the given loot
]]
function CLootMasterML:SetCandidateData( loot, candidate, name, value )
	local itemID = self:AddCandidate( loot, candidate )
	if not itemID then return end

	local rowID = self.lootTable[itemID]['candidates'][candidate]
	if not rowID then return end

	if not name then return end

	if name == 'response' then
		value = tonumber(value) or 1

		local responseData = CLootMaster.RESPONSE[value]
		if responseData then
			self:SetCandidateData( loot, candidate, 'responsesort', responseData.SORT, true )
		end
	end

	for index,data in pairs( self.lootTable[itemID]['rowdata'][rowID]['cols'] ) do
		if data and data['name'] == name then
			self.lootTable[itemID]['rowdata'][rowID]['cols'][index]['value'] = value
			return value
		end
	end

	self.lootTable[itemID]['rowdata'][rowID][name] = value

	return value
end

--[[
	Sets a candidate's response for a given item
]]
function CLootMasterML:SetCandidateResponse( link, candidate, response )
	response = tonumber(response) or 0
	local old = tonumber(self:GetCandidateData( link, candidate, 'response' ) or 0) or 0

	local loot = self:GetLoot(link)
	if not loot then return end

	if old<CLootMaster.RESPONSE.TIMEOUT and response >= CLootMaster.RESPONSE.TIMEOUT then
		-- Increate response count
		loot.numResponses = loot.numResponses + 1
	elseif old>=CLootMaster.RESPONSE.TIMEOUT and response < CLootMaster.RESPONSE.TIMEOUT then
		-- decreate response count
		loot.numResponses = loot.numResponses - 1
	end

	return self:SetCandidateData( link, candidate, 'response', response )
end

--[[
	Called when a candidate's response is set manually, informs the candidate of this action
]]
function CLootMasterML:SetManualResponse(loot, candidate, response)

	loot = self:GetLoot(loot)
	if not loot then return end

	local sResponse
	if CLootMaster.RESPONSE[response] then
		sResponse = CLootMaster.RESPONSE[response].TEXT
	end

	self:SetCandidateResponse(loot.id, candidate, response)
	self:SendWhisperResponse(format(L['Your selection for %s has been manually set to %s.'], loot.link or 'unknown item', sResponse or 'unknown'), candidate)

	if self:MonitorMessageRequired(loot.id) then
		self:SendMonitorMessage( 'SETCANDIDATEDATA', loot.id, candidate, 'response', response )
	end

	self:ReloadMLTableForLoot( loot.id )
end

--[[
	Looks up the candidates equipment for the slot by inspecting.
]]--
function CLootMasterML:InspectCandidate( loot, candidate )

	-- retrieve loot or return
	loot = self:GetLoot(loot)
	if not loot then return end

	-- See if candidate if human controlled... you never know ;)
	if not UnitPlayerControlled(candidate) then return end
	-- See if player is in inspection range.
	if not CheckInteractDistance(candidate,1) then return end

	-- inspect the candidate
	NotifyInspect(candidate)

	-- Retrieve the itemlink and levels for the equipSlot.
	local gear = CLootMaster:GetGearByINVTYPE(loot.equipLoc, candidate)

	local item1, item2 = strsplit('$', gear)
	self:SetCandidateData( loot.id, candidate, 'currentitem', item1, true )
	self:SetCandidateData( loot.id, candidate, 'currentitem2', item2, true )
	self:SetCandidateData( loot.id, candidate, 'foundGear', true, true )

	return true
end

--[[
	Retrieves some data in the candidate cache for the given loot
]]
function CLootMasterML:GetCandidateData( loot, candidate, name )

	if not loot then return end
	loot = self:GetLoot(loot)

	if not loot or not loot.id or not self.lootTable or not self.lootTable[loot.id] or not self.lootTable[loot.id]['candidates'] then return end

	local rowID = self.lootTable[loot.id]['candidates'][candidate]
	if not rowID then return end

	for index,value in pairs( self.lootTable[loot.id]['rowdata'][rowID]['cols'] ) do
		if value and value['name'] == name then
			return self.lootTable[loot.id]['rowdata'][rowID]['cols'][index]['value']
		end
	end

	return self.lootTable[loot.id]['rowdata'][rowID][name]
end

--[[
	Retrieves some data in the candidate cache for the given loot
]]
function CLootMasterML:GetCandidateDataByRowID( loot, rowID, name )

	if not loot then return end
	loot = self:GetLoot(loot)

	if not loot or not loot.id then return end
	if not rowID then return end

	for index,value in pairs( self.lootTable[loot.id]['rowdata'][rowID]['cols'] ) do
		if value and value['name'] == name then
			return self.lootTable[loot.id]['rowdata'][rowID]['cols'][index]['value']
		end
	end

	return self.lootTable[loot.id]['rowdata'][rowID][name]
end

--[[
	Adds an enchanter to the enchanting table.
]]
function CLootMasterML:AddEnchanter( enchanter )
	if self.enchanterTable['enchanters'][enchanter] then return end

	-- Find the guildRank and unlocalized classname from the guildCache
	local guildRankName = ''
	local enchanterClass = nil
	local enchanterClassLocalized = nil
	if self.guildInfo and self.guildInfo[enchanter] then
		enchanterClass = self.guildInfo[enchanter].class
		enchanterClassLocalized = self.guildInfo[enchanter].classLocalized
	end

	-- No class found, try looking it up another way.
	if not enchanterClass then
		enchanterClassLocalized, enchanterClass = UnitClass(enchanter)
		-- Update the reverse class lookup table
		if enchanterClassLocalized and enchanterClass then
			CLootMaster:UpdateClassLocalizer(enchanterClassLocalized, enchanterClass)
		end
	end

	tinsert(self.enchanterTable['rowdata'],
		{	["unitclass"]	= enchanterClass,
			["onclick"]		= function(arg1, button, down) if button=='LeftButton' then self:SetDisenchanter(enchanter, arg1) end end,
			["cols"] 		= {	{["name"]       = 'name',
								["value"]      = enchanter,
								["onenter"]    = addon.ShowCandidateCellPopup,
								["onleave"]    = addon.HideGearCellPopup,
								["color"]      = self.GetEnchanterClassCellColor,
								["colorargs"]  = { self, enchanter },
								["onenterargs"]= { self, enchanter },
								["onleaveargs"]= { self, enchanter }}
					}
		}
	)

	self.enchanterTable['enchanters'][enchanter] = #(self.enchanterTable['rowdata'])
end

--[[
	Retrieves data from the enchanter cache.
]]
function CLootMasterML:GetEnchanterData( enchanter, name )

	if not enchanter then return end

	local rowID = self.enchanterTable['enchanters'][enchanter]
	if not rowID then return end

	for index,value in pairs( self.enchanterTable['rowdata'][rowID]['cols'] ) do
		if value and value['name'] == name then
			return self.enchanterTable['rowdata'][rowID]['cols'][index]['value']
		end
	end

	return self.enchanterTable['rowdata'][rowID][name]

end


--[[
	Tries to give the loot to the given candidate
]]
function CLootMasterML:GiveLootToCandidate( link, candidate, lootType )

	local candidateID = nil
	local slotID = nil
	local loot = self:GetLoot(link)

	if not loot or not loot.id then
		return self:Print( format(L["Could not send %s to %s, loot not found in cache"], tostring(link), tostring(candidate) ) )
	end

	-- Look for the lootslotID
	for sID = 1, GetNumLootItems() do
		local sLink = GetLootSlotLink(sID)
		if sLink ~= nil then
			local sItemID = tostring(self:GetItemIDFromLink(sLink))
			local itemID = loot.id
			if sItemID and sItemID==itemID then
				slotID = sID
			break
			end
		end
	end

	if not slotID then
		return self:Print( format(L["Could not send %s to %s, lootslotID not found (already looted or lootwindow closed?) "], tostring(loot.link), tostring(candidate) ) )
	end

	-- Look for the candidateID
	for cID = 1, 40 do
		local cName = GetMasterLootCandidate(slotID, cID)
		if cName and cName==candidate then
			candidateID = cID
			break
		end
	end

	if not candidateID then
		return self:Print( format(L["Could not send %s to %s, candidate not found (offline, left group?)"], tostring(loot.link), tostring(candidate) ) )
	end



	self:SetCandidateData( loot.id, candidate, 'lootType', lootType or 0 )

	GiveMasterLoot( slotID, candidateID )

	local lootReason = ""

	if lootType==CLootMaster.LOOTTYPE.DISENCHANT then
		lootReason = " for disenchanting"
	end
	local lootMessage = format("%s went to %s%s", tostring(loot.link), tostring(candidate),lootReason)

	if CLootMaster:UnitInRaid('player') then
		SendChatMessage(lootMessage,"RAID")
	elseif UnitInParty('player') then
		SendChatMessage(lootMessage,"PARTY")
	end

end

--[[
	Called when the masterlooter changes, asks new masterlooter whether he wants
	to use this addon if that check is enabled, and reassigns current_ml
]]
function CLootMasterML:OnMasterLooterChange(masterlooter)
	-- if master looter is nil, return
	if not masterlooter then return end

	-- Is there really a new master looter?
	if self.current_ml and self.current_ml==masterlooter then return end

	-- cache the new ml
	self.current_ml = masterlooter

	-- if player is not the current master looter, then just return.
	if masterlooter~=UnitName('player') then return end

	-- Show a message here, based on the current settings
	if CLootMaster.db.profile.use_councillootmaster == 'enabled' then
		-- Always enable without asking
		CLootMaster:Print(L['you are the loot master, loot tracking enabled'])
		self:EnableTracking()
	elseif CLootMaster.db.profile.use_councillootmaster == 'disabled' then
		-- Disabled from the config panel
		CLootMaster:Print(L['you are the loot master, tracking disabled manually (configuration: /cloot config)'])
		self:DisableTracking()
	else
		LibDialog:Spawn("COUNCILLOOTMASTER_ASK_TRACKING")
	end
end

--[[
	Caches info about guild members for quick reference
]]
function CLootMasterML:CacheGuildInfo( event )
	local num = GetNumGuildMembers(false)
	for i=1, num do repeat
		local name, rank, rankIndex, _, classLocalized, _, _, _, _, _, classFileName = GetGuildRosterInfo(i)
		if name then
			local shortName, _ = strsplit("-", name)
			if not self.guildInfo[shortName] then
				self.guildInfo[shortName] = {}
			end
			self.guildInfo[shortName].rank = rank
			self.guildInfo[shortName].rankIndex = rankIndex
			self.guildInfo[shortName].class = classFileName
			self.guildInfo[shortName].classLocalized = classLocalized

			-- Update the reverse class lookup table
			CLootMaster:UpdateClassLocalizer(classLocalized, classFileName)
		end
	until true end
end



-- Deformat helps the parsing of the loot messages by creating a pattern from the
-- original string and caching this
local deformat_cache = {}
function CLootMasterML:Deformat(str, format)
local pat = deformat_cache[format]
if not pat then
	-- Escape special characters
	pat = format:gsub("[%^%$%(%)%%%.%[%]%*%+%-%?]",
						function(c) return "%"..c end)
	pat = pat:gsub("%%%%([sd])", {
							["s"] = "(.-)",
							["d"] = "(%d+)",
						})
	deformat_cache[format] = pat
end
return str:match(pat)
end

function CLootMasterML:DeformatC(format)
	return deformat_cache[format]
end


-- Used in FilterLootMessage(), just try each function and see if the 2nd return value is set
local LootMessageFilters = {
		function( lootMessage ) local l,c = CLootMasterML:Deformat( lootMessage, LOOT_ITEM_SELF_MULTIPLE ); return UnitName("player"),l,c end,
		function( lootMessage ) local l,c = CLootMasterML:Deformat( lootMessage, LOOT_ITEM_PUSHED_SELF_MULTIPLE ); return UnitName("player"),l,c end,
		function( lootMessage ) return CLootMasterML:Deformat( lootMessage, LOOT_ITEM_MULTIPLE ) end,
		function( lootMessage ) return UnitName("player"), CLootMasterML:Deformat( lootMessage, LOOT_ITEM_PUSHED_SELF ), 1 end,
		function( lootMessage ) return UnitName("player"), CLootMasterML:Deformat( lootMessage, LOOT_ITEM_SELF ), 1 end,
		function( lootMessage ) local p,l = CLootMasterML:Deformat( lootMessage, LOOT_ITEM ); return p,l,1 end
}

--[[
	Filters a lootmessage by reformatting the localised strings
	Returns playername, itemlink and itemcount.
]]
function CLootMasterML:FilterLootMessage( lootMessage )
	local player, link, count
	for i, func in pairs(LootMessageFilters) do
		player, link, count = func(lootMessage)
		if player and link and count then
			return player, link, count
		end
	end
	return false
end

--[[
	Check how your group looks when it is changed or you're just logging in
]]
function CLootMasterML:GROUP_UPDATE()
	currGroupSize = GetNumGroupMembers()
	if currGroupSize~=groupSize then
		groupSize = currGroupSize
		rosterNew = true
	end
	lootmethod, mlPartyID, mlRaidID = GetLootMethod()
	if lootmethod ~= 'master' then
		self.current_ml = nil
		return
	end
	if mlRaidID then
		-- we're in raid
		self:OnMasterLooterChange(GetRaidRosterInfo(mlRaidID))
	elseif mlPartyID==0 then
		-- player is ml
		self:OnMasterLooterChange(UnitName('player'))
	elseif mlPartyID then
		-- someone else in party is ml
		self:OnMasterLooterChange(UnitName('party'..mlPartyID))
	end
	if self.currentDE and not (CLootMaster:UnitInRaid(self.currentDE) or UnitInParty(self.currentDE)) then
		self.currentDE = nil
	end
end

--[[
	Someone received loot, lets see if it's in our cache. Send all candidates
	a message about this, so they can update their UI and register the lootdrop
	in a raidtracker for example
]]
function CLootMasterML:CHAT_MSG_LOOT( event, message )

	self:Debug("CHAT_MSG_LOOT")

	local sPlayer, sLink, iCount = self:FilterLootMessage(message)

	self:Debug(format("%s looted %sx%s", sPlayer or 'nil', sLink or 'nil', iCount or 'nil' ))

	if not sPlayer or sPlayer=='' or not sLink or sLink=='' or not iCount or tonumber(iCount)>1 then
		self:Debug("!player or !link or !iCount or iCount>1")
		return
	end

	-- ok, someone looted something... lets see if we can find it in our cache
	local loot = self:GetLoot( sLink )

	-- Did we find anything?
	if not loot or not loot.candidates then return self:Debug('loot not found') end

	-- Did we own this loot or are we just monitoring it?
	if not loot.mayDistribute then return self:Debug('only monitor loot, return') end

	-- We did, now get some info about the loot adressee
	local lootType = self:GetCandidateData( sLink, sPlayer, 'lootType' ) or CLootMaster.LOOTTYPE.UNKNOWN

	-- now send everyone in raid/party/candidates some info about the drop so they can update their ui
	if GetNumGroupMembers()>0 then
		self:Debug("send to raid")
		self:SendCommand( 'LOOTED', format('%s^%s^%s', sPlayer, sLink, lootType ), 'RAID' )
	elseif GetNumSubgroupMembers()>0 then
		self:Debug("send to party")
		self:SendCommand( 'LOOTED', format('%s^%s^%s', sPlayer, sLink, lootType ), 'PARTY' )
	else
		self:Debug("send to candidates")
		for candidate, id in pairs(loot.candidates) do
			self:SendCommand( 'LOOTED', format('%s^%s^%s', sPlayer, sLink, lootType ), candidate )
		end
	end

	-- Update the candidates status to LOOTED when we have more than 1 item
	if loot.quantity>1 then
		self:SetCandidateData( sLink, sPlayer, 'looted', true )
	end

	-- Since it's looted, remove it from the cache and update the UI
	self:RemoveLoot(sLink)
	self:UpdateUI()

	self:Debug("CHAT_MSG_LOOT end")

end

--[[
	Event triggers when a loot window is opened, ask enchanter skill of all party/raid members to generate
	a (new) list of disenchanters. If the loot window is opened the old list might be outdated, therefore we check/create again.
	actual list is created when handling responses command (command == 'ENCHRESP')
-]]
function CLootMasterML:LOOT_OPENED()
	if rosterNew then
		-- only ask enchanting skill when you are the master looter
		if self.current_ml==UnitName('player') then
			if CLootMaster:UnitInRaid('player') then
				self:SendCommand( 'ENCHSKILL', currentEnchNeeded, 'RAID' )
			elseif UnitInParty('player') and GetNumSubgroupMembers()>0 then
				self:SendCommand( 'ENCHSKILL', currentEnchNeeded, 'PARTY' )
			end
		end
		rosterNew = false
	end
end

--[[
	Event triggers when the master looter opens the popup on the loot screen
	In here we try to see what loot got selected and find the candidates for the
	loot. Ask the candidates if they want the loot. Tho, cache everything and make sure
	we don't ask the candidates more than one time (this could happen when the lootwindow
	gets opened more than one time)
]]
function CLootMasterML:OPEN_MASTER_LOOT_LIST()

	-- Check if CouncilLM needs to track the loot.
	if not self:TrackingEnabled() then return end

	-- Close the default confirm window
	StaticPopup_Hide("CONFIRM_LOOT_DISTRIBUTION")
	
	if IsAltKeyDown() and IsShiftKeyDown() then
		return
	end

	local _, lootName, lootQuantity, rarity = GetLootSlotInfo(LootFrame.selectedSlot)
	local link = GetLootSlotLink(LootFrame.selectedSlot)
	local itemID = self:GetItemIDFromLink(link)

	-- Traverse all lootslots and see how many of this item we have in total.
	local totalQuantity = 0
	local numLootSlots = GetNumLootItems()
	for slot=1, numLootSlots do
		local slotLink = GetLootSlotLink(slot)
		local slotItemID = self:GetItemIDFromLink(slotLink)
		if slotItemID and slotItemID==itemID then

			local _, _, slotQ, _ = GetLootSlotInfo(slot)

			-- A little sanity check; lets see if slotQuantity == 1
			if slotQ~=1 then
				self:Print( format(L["Could not redistribute %s because quantity != 1 (%s). Please handle it manually."], link, slotQ) )
				return
			end

			totalQuantity = totalQuantity + 1
		end

		self:Debug( format( 'OPEN_MASTER_LOOT_LIST: clicked %sx%s; slotCount %s; slotID %s; slotitem: %s', tostring(link), tostring(totalQuantity), tostring(numLootSlots), tostring(slot), tostring(slotLink) ) )
	end

	-- Fallback
	if not itemID then
		self:Print( format(L['Could not get itemcount for %s (no itemid found)'], tostring(link)) )
		totalQuantity = 1
	end

	-- Another sanity check... Check total quantity > 1
	if totalQuantity<1 then
		self:Print( format(L["Could not redistribute %s because total quantity < 1 (%s). Please handle it manually."], link, totalQuantity) )
		return
	end

	-- Addon is handling the loot, so lets close the default popup, unless alt is pressed
	if not IsAltKeyDown() then
		CloseDropDownMenus()
	end

	-- Check to see if we already have the loot registered
	if self:GetLootID(link) then
		-- loot is already registered, just update the ui and do nothing.
		local loot = self:GetLoot(link)
		loot.quantity = totalQuantity or 1
		self:ReloadMLTableForLoot( link )
		self:Debug( format('Updated %s quantity to %s', tostring(link), tostring(tonumber(totalQuantity)) ), true )
		return
	end

	-- Register the loot in the loottable
	local lootID = self:AddLoot(link, true, totalQuantity)
	if not lootID then return self:Print(L['Could not register loot']) end

	-- Auto announce?
	local autoAnnounce = rarity >= (CLootMaster.db.profile.auto_announce_threshold or 4)
	if CLootMaster.db.profile.auto_announce_threshold == 0 then autoAnnounce=false end

	local classResult = {} -- table with true/false for eligible classes
	
	-- Ok Lets see. Who are the candidates for this slot?
	if IsAltKeyDown() then
		for candidateID = 1, 40 do repeat
			local candidate = GetMasterLootCandidate(LootFrame.selectedSlot, candidateID)
			if not candidate then break end
		
			if not self:IsCandidate( lootID, candidate ) then
				-- Create the candidate for link
				self:AddCandidate( lootID, candidate )
			end
		until true end
		classResult = CLootMaster:FullClassesEncoded()
		self:Print('Adding all members to candidate list (alt+click detected)')
	else
		for candidateID = 1, 40 do repeat
			local candidate = GetMasterLootCandidate(LootFrame.selectedSlot, candidateID)
			if not candidate then break end
			
			local playerClass = nil
			local playerLocalClass = nil
			if self.guildInfo and self.guildInfo[candidate] then
				playerClass = self.guildInfo[candidate].class
				playerLocalClass = self.guildInfo[candidate].classLocalized
			end

			-- No class found, try looking it up another way.
			if not playerClass then
				playerLocalClass, playerClass = UnitClass(candidate)
				-- Update the reverse class lookup table
				if playerLocalClass and playerClass then
					CLootMaster:UpdateClassLocalizer(playerLocalClass, playerClass)
				end
			end
		
			if not self:IsCandidate( lootID, candidate ) and playerClass ~= nil then
				if classResult[playerClass] then
					-- Create the candidate for link
					self:AddCandidate( lootID, candidate )
				elseif classResult[playerClass]==nil then
					classResult[playerClass] = CLootMaster:EligibleItemCheck( link, playerLocalClass, playerClass )
					if classResult[playerClass] then
						self:AddCandidate( lootID, candidate )
					end
				end
			end
		until true end
		classResult = CLootMaster:EncodeUnlocalizedClasses(classResult)
	end

	-- Set the loot status to not announced.
	self.lootTable[lootID].announced = false
	
	-- Set the eligible classes
	self.lootTable[lootID].encodedClasses = classResult

	-- Lets see if we have to autoloot
	local isAutoLooted = false
	if CLootMaster.db.profile.AutoLootThreshold~=0 and CLootMaster.db.profile.AutoLooter~='' and self.lootTable[lootID].autoLootable then
		-- loot is below or equal to AutoLootThreshold and matches the autoLooter requirements
		-- try to give the loot.
		autoAnnounce = false

		if IsAltKeyDown() then
			self:Print(L['Not auto looting (alt+click detected)'])
		else
			isAutoLooted = true
			if self:IsCandidate( lootID, CLootMaster.db.profile.AutoLooter or '' ) then
				self:Print(format(L['Auto looting %s to %s'], link or 'nil', CLootMaster.db.profile.AutoLooter or 'nil'))
				-- dont know if it will ever happen, but send all matching items to the autolooter
				for i=1, totalQuantity do
					self:GiveLootToCandidate( lootID, CLootMaster.db.profile.AutoLooter or '', CLootMaster.LOOTTYPE.DISENCHANT )
				end
			else
				self:Print(format(L['Auto looting of %s to %s failed. Not a candidate for this loot.'], link or 'nil', CLootMaster.db.profile.AutoLooter or 'nil'))
			end
		end
	end

	-- See if we have to auto announce
	if autoAnnounce then
		if IsAltKeyDown() then
			self:Print(L['Not auto announcing (alt+click detected)'])
		else
			self:AnnounceLoot(lootID)
		end
	end

	-- Update the UI
	self:ReloadMLTableForLoot( lootID )

	-- Send candidate list to monitors
	if self:MonitorMessageRequired( lootID ) then
		self:SendCandidateListToMonitors( lootID )
	end

end